"""Main module."""
import json
import os
import requests

import json, os, requests
import urllib3
import datetime
import time

from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

import logging

from biolmai.const import ACCESS_TOK_PATH

log = logging.getLogger('biolm_util')


def requests_retry_session(
    retries=3,
    backoff_factor=0.3,
    status_forcelist=list(range(400, 599)),
    session=None,
):
    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session


def retry_minutes(sess, URL, HEADERS, dat, timeout, mins):
    """Retry for N minutes."""
    try:
        now = datetime.datetime.now()
        try_until = now + datetime.timedelta(minutes=mins)
        while datetime.datetime.now() < try_until:
            response = None
            try:
                log.info('Trying {}'.format(datetime.datetime.now()))
                response = sess.post(
                    URL,
                    headers=HEADERS,
                    data=dat,
                    timeout=timeout
                )
                response.raise_for_status()
                if 'error' in response.json():
                    raise ValueError(response.json().dumps())
                else:
                    break
            except Exception as e:
                log.warning(e)
                if response:
                    log.warning(response.text)
                time.sleep(5)  # Wait 5 seconds between tries
        if response is None:
            err = "Got Nonetype response"
            raise ValueError(err)
        elif 'Server Error' in response.text:
            err = "Got Server Error"
            raise ValueError(err)
        else:
            response.raise_for_status()
    except Exception as e:
        raise
    else:
        return response


def get_user_auth_header():
    """Returns a dict with the appropriate Authorization header, either using
    an API token from BIOLMAI_TOKEN environment variable, or by reading the
    credentials file at ~/.biolmai/credntials next."""
    api_token = os.environ.get('BIOLMAI_TOKEN', None)
    if api_token:
        headers = {'Authorization': f'Token {api_token}'}
    elif os.path.exists(ACCESS_TOK_PATH):
        with open(ACCESS_TOK_PATH, 'r') as f:
            access_refresh_dict = json.load(f)
        access = access_refresh_dict.get('access')
        refresh = access_refresh_dict.get('refresh')
        headers = {
            'Cookie': 'access={};refresh={}'.format(access, refresh),
            'Content-Type': 'application/json'
        }
    else:
        AssertionError("No https://biolm.ai credentials found. Please run "
                       "`biolmai status` to debug.")
    return headers


def get_api_token():
    """Get a BioLM API token to use with future API requests.

    Copied from https://api.biolm.ai/#d7f87dfd-321f-45ae-99b6-eb203519ddeb.
    """
    url = "https://biolm.ai/api/auth/token/"

    payload = json.dumps({
        "username": os.environ.get("BIOLM_USER"),
        "password": os.environ.get("BIOLM_PASSWORD")
    })
    headers = {
        'Content-Type': 'application/json'
    }

    response = requests.request("POST", url, headers=headers, data=payload)
    response_json = response.json()

    return response_json


def api_call(model_name, action, payload, access, refresh):
    """Hit an arbitrary BioLM model inference API."""
    # Normally would POST multiple sequences at once for greater efficiency,
    # but for simplicity sake will do one at at time right now
    url = f'https://biolm.ai/api/v1/models/{model_name}/{action}/'

    payload = json.dumps(payload)

    try:
        assert access
        assert refresh
    except AssertionError:
        raise AssertionError("BioLM access or refresh token not set")

    headers = {
        'Cookie': 'access={};refresh={}'.format(access, refresh),
        'Content-Type': 'application/json'
    }

    session = requests_retry_session()
    tout = urllib3.util.Timeout(total=180, read=180)
    response = retry_minutes(session, url, headers, payload, tout, mins=12)

    resp_json = response.json()

    return resp_json



import os

if os.environ.get('BIOLMAI_LOCAL', False):
    BASE_DOMAIN = 'http://localhost:8000'
else:
    BASE_DOMAIN = 'https://biolm.ai'

USER_BIOLM_DIR = os.path.join(os.path.expanduser('~'), '.biolmai')
ACCESS_TOK_PATH = os.path.join(USER_BIOLM_DIR, 'credentials')
GEN_TOKEN_URL = f'{BASE_DOMAIN}/ui/accounts/user-api-tokens/'



"""Console script for biolmai."""
import sys
import click
import os
from biolmai.auth import generate_access_token, \
    get_auth_status, save_access_refresh_token
from biolmai.const import ACCESS_TOK_PATH


@click.command()
def main(args=None):
    """Console script for biolmai."""
    click.echo("Replace this message by putting your code into "
               "biolmai.cli.main")
    click.echo("See click documentation at https://click.palletsprojects.com/")
    return 0


@click.group()
@click.option('--debug/--no-debug', default=False)
def cli(debug):
    pass


@cli.command()  # @cli, not @click!
def status():
    get_auth_status()


@cli.command()
def login():
    uname = click.prompt("Username", default=None, hide_input=False,
                  confirmation_prompt=False, type=str)
    password = click.prompt("Password", default=None, hide_input=True,
                  confirmation_prompt=False, type=str)
    access_refresh_tok_dict = generate_access_token(uname, password)
    try:
        access = access_refresh_tok_dict.get('access')
        refresh = access_refresh_tok_dict.get('refresh')
        click.echo("Saving new access and refresh token.")
        save_access_refresh_token(access_refresh_tok_dict)
    except Exception as e:
        click.echo("Unhandled login exception!")
        raise


@cli.command()
def logout():
    os.remove(ACCESS_TOK_PATH)


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



"""Top-level package for BioLM AI."""
__author__ = """Nikhil Haas"""
__email__ = 'nikhil@biolm.ai'
__version__ = '0.1.1'

from biolmai.biolmai import get_api_token, api_call


__all__ = [
    "get_api_token",
    "api_call",
]



#!/usr/bin/env python

"""Tests for `biolmai` package."""

import pytest

from click.testing import CliRunner

import biolmai
from biolmai import cli

import logging
log = logging.getLogger(__name__)


def test_authentication():
    """Test to make sure the environment variables for auth work, and
    that you get tokens back from the site to use for requests."""
    resp = biolmai.get_api_token()
    # Make sure we have access and refresh keys in dictionary response
    assert 'access' in resp, log.warning(resp)
    assert 'refresh' in resp, log.warning(resp)


def test_arbitrary_api_call_inference():
    seq = "MSILVTRPSPAGEELVSRLRTLGQVAWHFPLIEFSPGQQLPQLADQLAALGESDLLFALSQHAVAFA"
    payload = {
        "instances": [{
            "data": {"text": seq}
        }]
    }
    tokens = biolmai.get_api_token()
    access = tokens.get('access')
    refresh = tokens.get('refresh')

    resp = biolmai.api_call(
        model_name='esmfold-singlechain',
        action='predict',
        payload=payload,
        access=access,
        refresh=refresh
    )

    assert 'predictions' in resp, log.warning(resp)


@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.main)
    assert result.exit_code == 0
    assert 'biolmai.cli.main' in result.output
    help_result = runner.invoke(cli.main, ['--help'])
    assert help_result.exit_code == 0
    assert '--help  Show this message and exit.' in help_result.output



import json
import os
import pprint
import stat

import click
import requests

from biolmai.const import ACCESS_TOK_PATH, BASE_DOMAIN, GEN_TOKEN_URL, \
    USER_BIOLM_DIR


def validate_user_auth(api_token=None, access=None, refresh=None):
    """Validates an API token, to be used as 'Authorization: Token 1235abc'
    authentication method."""
    url = f'{BASE_DOMAIN}/api/v1/auth/login-check/'
    if api_token is not None:
        headers = {'Authorization': f'Token {api_token}'}
    else:
        headers = {
            'Cookie': 'access={};refresh={}'.format(access, refresh),
            'Content-Type': 'application/json'
        }
    try:
        # click.echo(headers)
        r = requests.post(url=url, headers=headers)
        # click.echo('Printing')
        # click.echo(r.text)
        json_response = r.json()
        pretty_json = pprint.pformat(json_response, indent=2)
        click.echo(pretty_json)
    except Exception:
        click.echo("Token validation failed!\n")
        raise
    else:
        return r


def refresh_access_token(refresh):
    """Attempt to refresh temporary user access token, by using their refresh
    token, which has a longer TTL."""
    url = f'{BASE_DOMAIN}/api/auth/token/refresh/'
    headers = {
        'Cookie': 'refresh={}'.format(refresh),
        'Content-Type': 'application/json'
    }
    r = requests.post(url=url, headers=headers)
    json_response = r.json()
    if r.status_code != 200:
        pretty_json = pprint.pformat(json_response, indent=2)
        click.echo(pretty_json)
        click.echo(f"Token refresh failed! Please login by "
                   "running `biolmai login`.\n")
    else:
        click.echo("User access token successfully refreshed. Saving credentials...")
        access_refresh_dict = {'access': json_response['access'],
                               'refresh': refresh}
        save_access_refresh_token(access_refresh_dict)


def get_auth_status():
    environ_token = os.environ.get('BIOLMAI_TOKEN', None)
    if environ_token:
        msg = "Environment variable BIOLMAI_TOKEN detected. Validating token..."
        click.echo(msg)
        validate_user_auth(api_token=environ_token)
    elif os.path.exists(ACCESS_TOK_PATH):
        msg = f"Credentials file found {ACCESS_TOK_PATH}. Validating token..."
        click.echo(msg)
        with open(ACCESS_TOK_PATH, 'r') as f:
            access_refresh_dict = json.load(f)
        access = access_refresh_dict.get('access')
        refresh = access_refresh_dict.get('refresh')
        resp = validate_user_auth(access=access, refresh=refresh)
        if resp.status_code != 200:
            click.echo("Access token validation failed. Attempting to refresh token...")
            # Attempt to use the 'refresh' token to get a new 'access' token
            refresh_access_token(refresh)
    else:
        msg = f"No https://biolm.ai login credentials found. Please " \
              f"set the environment variable BIOLMAI_TOKEN to a token from {GEN_TOKEN_URL}, or login by " \
              "running `biolmai login`."
        click.echo(msg)


def generate_access_token(uname, password):
    """Generate a TTL-expiry access and refresh token, to be used as
    'Cookie: acccess=; refresh=;" headers, or the access token only as a
    'Authorization: Bearer 1235abc' token.

    The refresh token will expire in hours or days, while the access token
    will have a shorter TTL, more like hours. Meaning, this method will
    require periodically re-logging in, due to the token expiration time. For a
    more permanent auth method for the API, use an API token by setting the
    BIOLMAI_TOKEN environment variable.
    """
    url = f'{BASE_DOMAIN}/api/auth/token/'
    try:
        r = requests.post(url=url, data={'username': uname, 'password': password})
        json_response = r.json()
    except Exception:
        click.echo("Login failed!\n")
        raise
    if r.status_code != 200:
        click.echo("Login failed!\n")
        resp_json = r.json()
        pretty_json = pprint.pformat(resp_json, indent=2)
        click.echo(pretty_json)
        return {}
    else:
        click.echo("Login succeeded!\n")
        return json_response


def save_access_refresh_token(access_refresh_dict):
    """Save temporary access and refresh tokens to user folder for future
    use."""
    os.makedirs(USER_BIOLM_DIR, exist_ok=True)
    # Save token
    with open(ACCESS_TOK_PATH, 'w') as f:
        json.dump(access_refresh_dict, f)
    os.chmod(ACCESS_TOK_PATH, stat.S_IRUSR | stat.S_IWUSR)
    # Validate token and print user info
    # click.echo(access_refresh_dict)
    access = access_refresh_dict.get('access')
    refresh = access_refresh_dict.get('refresh')
    validate_user_auth(access=access, refresh=refresh)
